// Copyright (C) 2022  Aravinth Manivannan <realaravinth@batsense.net>
// SPDX-FileCopyrightText: 2023 Aravinth Manivannan <realaravinth@batsense.net>
//
// SPDX-License-Identifier: AGPL-3.0-or-later

use std::env;
use std::path::PathBuf;
use std::sync::Arc;

use actix_web::{
    error::InternalError, http::StatusCode, middleware as actix_middleware, web::Data as WebData,
    web::JsonConfig, App, HttpServer,
};
use clap::{Args, Parser, Subcommand, ValueEnum};
//use static_assets::FileMap;
use tracing::info;
use tracing_actix_web::TracingLogger;
//
//pub use crate::api::v1::ROUTES as V1_API_ROUTES;
use ctx::CliCtx;
use ctx::DaemonCtx;
use ctx::MinAppContext;
pub use settings::Settings;

pub const CACHE_AGE: u32 = 604800;

pub const GIT_COMMIT_HASH: &str = env!("GIT_HASH");
pub const VERSION: &str = env!("CARGO_PKG_VERSION");
pub const PKG_NAME: &str = env!("CARGO_PKG_NAME");
pub const PKG_DESCRIPTION: &str = env!("CARGO_PKG_DESCRIPTION");
pub const PKG_HOMEPAGE: &str = env!("CARGO_PKG_HOMEPAGE");

pub type AppFullCtx = WebData<ctx::ArcFullCtx>;
pub type AppMinCtx = WebData<ctx::ArcMinCtx>;

mod api;
mod complaince;
mod ctx;
mod db;
mod docker;
mod docker_compose;
mod errors;
mod git;
mod pages;
mod runner;
mod settings;
mod utils;

//lazy_static::lazy_static! {
//    pub static ref FILES: FileMap = FileMap::new();
//}
//
//#[derive(Parser)]
//#[clap(author, version, about, long_about = None)]
#[derive(Parser)]
#[clap(author, version, about, long_about = None)]
struct Cli {
    #[command(subcommand)]
    command: Command,
}

#[derive(Subcommand)]
enum Command {
    Daemon,
    Verify { path: PathBuf },
    Test { path: PathBuf },
}

#[actix_web::main]
#[cfg(not(tarpaulin_include))]
async fn main() -> std::io::Result<()> {
    if env::var("RUST_LOG").is_err() {
        env::set_var("RUST_LOG", "info");
    }

    pretty_env_logger::init();
    let cli = Cli::parse();

    info!(
        "{}: {}.\nFor more information, see: {}\nBuild info:\nVersion: {} commit: {}",
        PKG_NAME, PKG_DESCRIPTION, PKG_HOMEPAGE, VERSION, GIT_COMMIT_HASH
    );

    match cli.command {
        Command::Daemon => run_daemon().await,
        Command::Verify { path } => {
            let s = std::fs::read_to_string(path).unwrap();
            let _: complaince::target::Target = toml::from_str(&s).unwrap();
            println!("Syntax OK");
            Ok(())
        }
        Command::Test { path } => {
            let ctx: Arc<dyn MinAppContext> = Arc::new(CliCtx::new());
            let serv = basic_server(ctx.clone()).await;

            loop {
                log::info!("Waiting for server to start...");
                let res = reqwest::get(&format!(
                    "http://localhost:29130{}",
                    api::v1::meta::META.build_details
                ))
                .await;
                if res.is_ok() {
                    log::info!("Waiting server started");
                    break;
                }
                tokio::time::sleep(std::time::Duration::new(2, 0)).await;
            }

            crate::runner::suite::SuiteRunnerState::run_proxy(ctx.as_ref());
            let (suite_results, init_containers, specimen_logs) =
                crate::runner::target::run_target(ctx.as_ref(), path.clone()).await;
            let content = crate::runner::results::ArchivableResult {
                commit: "".into(),
                suites: suite_results,
                init_containers,
               specimen_logs,
            };
            let results_file = path.join("results.json");
            println!("Writing results to: {:?}", path.canonicalize());
            std::fs::write(results_file, serde_json::to_string(&content).unwrap()).unwrap();
            let results_html_file = path.join("results.html");
            let contents = pages::ViewResult::new(pages::Payload { results: content }).render();
            std::fs::write(results_html_file, contents).unwrap();
            println!("Writing results to: {:?}", path.canonicalize());

            crate::runner::suite::SuiteRunnerState::stop_proxy(ctx.as_ref());

            serv.stop(true).await;
            Ok(())
        }
    }
}

async fn run_daemon() -> std::io::Result<()> {
    let settings = Settings::new().unwrap();
    settings.init();
    let inner_ctx = Arc::new(DaemonCtx::new(settings.clone()).await);
    let ctx = AppFullCtx::new(inner_ctx.clone());
    let ctx2 = AppMinCtx::new(inner_ctx);
    ctx.db().migrate().await.unwrap();

    if ctx.settings().docker_proxy {
        crate::runner::suite::SuiteRunnerState::run_proxy(ctx.as_ref());
    }

    daemon(ctx.clone(), ctx2).await?;
    if ctx.settings().docker_proxy {
        crate::runner::suite::SuiteRunnerState::stop_proxy(ctx.as_ref());
    }
    Ok(())
}

async fn basic_server(ctx: Arc<dyn MinAppContext>) -> actix_web::dev::ServerHandle {
    let ctx = AppMinCtx::new(ctx);
    info!("Starting server on: http://0.0.0.0:29130");
    let serv = HttpServer::new(move || {
        App::new()
            .wrap(TracingLogger::default())
            .wrap(actix_middleware::Compress::default())
            .app_data(ctx.clone())
            .app_data(get_json_err())
            .wrap(
                actix_middleware::DefaultHeaders::new()
                    .add(("Permissions-Policy", "interest-cohort=()")),
            )
            .wrap(actix_middleware::NormalizePath::new(
                actix_middleware::TrailingSlash::Trim,
            ))
            .configure(services)
    })
    .bind("0.0.0.0:29130")
    .unwrap()
    .run();
    let handle = serv.handle();
    tokio::spawn(serv);
    handle
}

async fn daemon(ctx: AppFullCtx, ctx2: AppMinCtx) -> std::io::Result<()> {
    let ip = ctx.settings().server.get_ip();
    let workers = ctx.settings().server.workers.unwrap_or_else(num_cpus::get);
    let scheduler = runner::scheduler::Scheduler::spawn(ctx.clone()).await;

    info!("Starting server on: http://{}", ip);
    HttpServer::new(move || {
        App::new()
            .wrap(TracingLogger::default())
            .wrap(actix_middleware::Compress::default())
            .app_data(ctx.clone())
            .app_data(ctx2.clone())
            .app_data(get_json_err())
            .wrap(
                actix_middleware::DefaultHeaders::new()
                    .add(("Permissions-Policy", "interest-cohort=()")),
            )
            .wrap(actix_middleware::NormalizePath::new(
                actix_middleware::TrailingSlash::Trim,
            ))
            .configure(services)
    })
    .workers(workers)
    .bind(ip)
    .unwrap()
    .run()
    .await?;
    info!("Stopping job runner");
    scheduler.stop().await;
    Ok(())
}

#[cfg(not(tarpaulin_include))]
pub fn get_json_err() -> JsonConfig {
    JsonConfig::default().error_handler(|err, _| {
        //debug!("JSON deserialization error: {:?}", &err);
        InternalError::new(err, StatusCode::BAD_REQUEST).into()
    })
}

pub fn services(cfg: &mut actix_web::web::ServiceConfig) {
    crate::api::v1::services(cfg);
}
